import torch
import argparse
import cv2
import numpy as np
import time
import os
import utils.custom_transforms as transformer
from utils.patch_transformer import Adversarial_Patch
from models.adversarial_models import AdversarialModels
from utils.dataloader import LoadFromImageFile
from utils.utils import makedirs, to_cuda_vars, format_time
from tqdm import tqdm
import warnings
warnings.simplefilter('ignore')

parser = argparse.ArgumentParser(description='Generating Adversarial Patches')
parser.add_argument('--data_root', type=str, help='path to dataset', default='Src/input_img')
parser.add_argument('--train_list', type=str, default='Src/list/test_list.txt')
parser.add_argument('--print_file', type=str, default='Src/list/printable30values.txt')
parser.add_argument('--distill_ckpt', type=str, default="models/guo/distill_model.ckpt")
parser.add_argument('--height', type=int, help='input image height', default=256)
parser.add_argument('--width', type=int, help='input image width', default=512)
parser.add_argument('-b', '--batch_size', type=int, help='mini-batch size', default=2)
parser.add_argument('-j', '--num_threads', type=int, help='data loading workers', default=0)
parser.add_argument('--lr', type=float, help='initial learning rate', default=1e-3)
parser.add_argument('--num_epochs', type=int, help='number of total epochs', default=40)
parser.add_argument('--seed', type=int, help='seed for random functions, and network initialization', default=0)
parser.add_argument('--patch_size', type=int, help='Resolution of patch', default=256)
parser.add_argument('--patch_shape', type=str, help='circle or square', default='circle')
parser.add_argument('--patch_path', type=str, help='Initialize patch from file')
parser.add_argument('--mask_path', type=str, help='Initialize mask from file')
parser.add_argument('--target_disp', type=int, default=120)
parser.add_argument('--model', nargs='*', type=str, default='distill', choices=['distill'], help='Model architecture')
parser.add_argument('--name', type=str, help='output directory', default="result")
args = parser.parse_args()


def main():
    save_path = 'Dst/checkpoints/' + args.name
    print('===============================')
    print('=> Everything will be saved to \"{}\"'.format(save_path))
    makedirs(save_path)

    torch.manual_seed(args.seed)
    torch.backends.cudnn.benchmark = True

    # DataLoader
    train_transform = transformer.Compose([
        transformer.RandomHorizontalFlip(),
        transformer.RandomAugumentColor(),
        transformer.RandomScaleCrop(scale_range=(0.85, 1.0)),
        transformer.ResizeImage(h=args.height, w=args.width),
        transformer.ArrayToTensor()
    ])

    train_set = LoadFromImageFile(
        args.data_root,
        args.train_list,
        seed=args.seed,
        train=True,
        monocular=True,
        transform=train_transform,
        extension=".jpg"
    )

    train_loader = torch.utils.data.DataLoader(
        dataset=train_set,
        batch_size=args.batch_size,
        shuffle=True,
        num_workers=args.num_threads,
        pin_memory=True,
        drop_last=True
    )

    print('===============================')
    # Attacked Models
    models = AdversarialModels(args)
    models.load_ckpt()

    # Patch and Mask
    adv = Adversarial_Patch(
        patch_type=args.patch_shape,
        batch_size=args.batch_size,
        image_size=min(args.width, args.height),
        patch_size=args.patch_size,
        printability_file=args.print_file
    )

    if not args.patch_path:
        patch_cpu, mask_cpu = adv.initialize_patch_and_mask()
    else:
        patch_cpu, mask_cpu = adv.load_patch_and_mask_from_file(args.patch_path, args.mask_path, npy=True)
    patch_cpu.requires_grad_(True)

    # Optimizer
    optimizer = torch.optim.Adam([patch_cpu], lr=args.lr, amsgrad=True)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', patience=50)

    # Train
    print('===============================')
    print("Start training ...")
    start_time = time.time()
    for epoch in range(args.num_epochs):
        ep_nps_loss, ep_tv_loss, ep_loss, ep_disp_loss = 0, 0, 0, 0
        ep_time = time.time()

        for i_batch, sample in tqdm(enumerate(train_loader), desc=f'Running epoch {epoch}', total=len(train_loader), leave=False):
            with torch.autograd.detect_anomaly():
                sample = to_cuda_vars(sample)  # send item to gpu
                sample.update(models.get_original_disp(sample))  # get non-attacked disparity

                img, original_disp = sample['left'], sample['original_distill_disp']
                patch, mask = patch_cpu.cuda(), mask_cpu.cuda()

                # transform patch
                patch_t, mask_t = adv.patch_transformer(
                    patch=patch,
                    mask=mask,
                    batch_size=args.batch_size,
                    img_size=min(args.width, args.height),
                    do_rotate=True,
                    rand_loc=True,
                    random_size=True,
                    train=True
                )

                # apply transformed patch to clean image
                attacked_img = adv.patch_applier(img, patch_t, mask_t)

                # Loss
                disp_loss = 0

                if 'distill' in args.model:
                    adv_target_distill_disp = adv.create_fake_disp(original_disp.detach(), mask_t[:, 0, :, :].detach(), args.target_disp)
                    est_disp = models.distill(attacked_img)
                    distill_loss = torch.nn.functional.l1_loss(torch.mul(mask_t, adv_target_distill_disp), torch.mul(mask_t, est_disp)).mean()
                    disp_loss += distill_loss

                nps_loss = adv.calculate_nps(patch)
                tv_loss = adv.calculate_tv(patch)

                loss = disp_loss + nps_loss * 0.01 + torch.max(tv_loss, torch.tensor(0.1).cuda()) * 2.5

                ep_nps_loss += nps_loss.detach().cpu().numpy()
                ep_tv_loss += tv_loss.detach().cpu().numpy()
                ep_disp_loss += disp_loss.detach().cpu().numpy()
                ep_loss += loss.detach().cpu().numpy()

                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                models.distill.zero_grad()

                patch_cpu.data.clamp_(0, 1)  # keep patch in image range

                del patch_t, loss, nps_loss, tv_loss, disp_loss
                torch.cuda.empty_cache()

        ep_disp_loss = ep_disp_loss/len(train_loader)
        ep_nps_loss = ep_nps_loss/len(train_loader)
        ep_tv_loss = ep_tv_loss/len(train_loader)
        ep_loss = ep_loss/len(train_loader)
        scheduler.step(ep_loss)

        ep_time = time.time() - ep_time
        total_time = time.time() - start_time
        print('===============================')
        print(' FIN EPOCH: ', epoch)
        print('TOTAL TIME: ', format_time(int(total_time)))
        print('EPOCH TIME: ', format_time(int(ep_time)))
        print('EPOCH LOSS: ', ep_loss)
        print(' DISP LOSS: ', ep_disp_loss)
        print('  NPS LOSS: ', ep_nps_loss)
        print('   TV LOSS: ', ep_tv_loss)
        np.save(save_path + '/epoch_{}_patch.npy'.format(str(epoch)), patch_cpu.data.numpy())
        np.save(save_path + '/epoch_{}_mask.npy'.format(str(epoch)), mask_cpu.data.numpy())


if __name__ == '__main__':
    main()
